/* wptClipEditDlg.cpp - Clipboard Editor dialog
 *	Copyright (C) 2000-2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>

#include "wptTypes.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptErrors.h"
#include "wptGPG.h"
#include "wptNLS.h"
#include "resource.h"

#ifdef _MSC_VER
#include "winpt_header.h"

static DWORD help_arr[] = {
    IDC_CLIPEDIT_SEND,	WPT_CLIPEDIT_COPY,
    IDC_CLIPEDIT_CLEAR,	WPT_CLIPEDIT_CLEAR,
    IDC_CLIPEDIT_LOAD,	WPT_CLIPEDIT_LOAD,
    IDC_CLIPEDIT_SAVE,	WPT_CLIPEDIT_SAVE,
    IDC_CLIPEDIT_QUOTE,	WPT_CLIPEDIT_QUOTE,
    0, 0};
#endif

/* Load the clipboard contents into the text field in @dlg.
   Return value: 0 on success. */
static int
load_clipboard (HWND dlg)
{
    HANDLE clipmem;
    char *cliptext;
    char *p;

    CloseClipboard (); /* make sure it is closed. */
    if (OpenClipboard (NULL) == FALSE) {
	msg_box (dlg, winpt_strerror (WPTERR_CLIP_OPEN), _("Clipboard"), MB_ERR);
	return -1;
    }
    clipmem = GetClipboardData (CF_TEXT);
    if (clipmem == NULL) {
	CloseClipboard ();
	return -1;
    }
    
    cliptext = (char *) GlobalLock (clipmem);
    if (cliptext == NULL || !strlen (cliptext)) {    
	CloseClipboard ();
	return -1;
    }

    p = new char [strlen (cliptext)+2];
    if (!p)    
	BUG (0);	
    strcpy (p, cliptext);
    
    GlobalUnlock (clipmem);
    CloseClipboard ();	
    SetDlgItemText (dlg, IDC_CLIPEDIT_CLIPTEXT2, p);	
    free_if_alloc (p);
    
    return 0;
}


/* Load the clipboard from the selected file. */
static int
load_clipboard_from_file (HWND dlg)
{
    OPENFILENAME open;
    FILE *fp;
    char *p;
    char file[300] = "";
    DWORD size;
    int id;

    memset (&open, 0, sizeof (open));
    open.lStructSize = sizeof (OPENFILENAME);	    
    open.hInstance = glob_hinst;	    
    open.lpstrTitle = _("File Open");	    
    open.lpstrFilter = (char *)"All Files (*.*)\0*.*\0\0";
    open.hwndOwner = dlg;	    
    open.lpstrFile = file;		    
    open.nMaxFile = sizeof (file)-1;		    
    open.Flags = OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST;
	    
    if (GetOpenFileName (&open)) {
	if( (size=get_file_size (file)) > MAX_CLIPTEXT_SIZE) {	    
	    id = msg_box (dlg, _("The file you want to add is very large.\n"
			         "Still proceed?"), _("Clipboard"), 
			  MB_INFO|MB_YESNO);
	    if (id == IDNO)
		return -1;
	}
	fp = fopen (file, "rb");
	if (!fp) {
	    msg_box (dlg, winpt_strerror (WPTERR_FILE_OPEN), 
		     _("Clipboard"), MB_ERR);
	    return FALSE;
	}
	p = new char[size+1];
	if (!p)
	    BUG (0);
	fread (p, 1, size, fp);
	p[size] = '\0';
	fclose (fp);
	SetDlgItemText (dlg, IDC_CLIPEDIT_CLIPTEXT2, p);	
	free_if_alloc (p);
    }
    return 0;
}


/* Save the contents of the clipboard to the selected file. */
static int
save_clipboard_to_file (HWND dlg)
{
    OPENFILENAME save;
    FILE *fp;
    char *p=NULL;
    char file[300] = "";
    DWORD nbytes;
    int id;

    memset (&save, 0, sizeof (save));	    
    save.lStructSize = sizeof (OPENFILENAME);	    
    save.hInstance = glob_hinst;	    
    save.lpstrTitle = _("File Save");	    
    save.lpstrFile = (char *)"All Files (*.*)\0*.*\0\0";
    save.hwndOwner = dlg;		    
    save.lpstrFile = file;		    
    save.nMaxFile = sizeof (file) - 1;
    save.Flags = OFN_OVERWRITEPROMPT;
            	    
    if( GetSaveFileName( &save ) ) {
	if( file_exist_check( file ) == 0 ) {	    
	    id = log_box (_("Clipboard"), MB_YESNO,
			  _("\"%s\" already exists.\n"
			    "Replace existing file?"), file);	    
	    if( id == IDNO )	
		return FALSE;
	}
	fp = fopen (file, "wb");
	if (!fp) {
	    msg_box (dlg, winpt_strerror (WPTERR_FILE_CREAT), 
		     _("Clipboard"), MB_ERR);
	    return FALSE;
	}
	/* XXX: there is a NUL-byte at the end. */
	nbytes = SendDlgItemMessage (dlg, IDC_CLIPEDIT_CLIPTEXT2, 
				     WM_GETTEXTLENGTH, 0, 0);
	if (nbytes > 0) {
	    p = new char[nbytes+1];
	    if (!p)
		BUG (NULL);
	    GetDlgItemText (dlg, IDC_CLIPEDIT_CLIPTEXT2, p, nbytes);
	    fwrite (p, 1, nbytes, fp);
	    
	}	   
	fclose (fp);
	free_if_alloc (p);
    }
    return 0;
}


/* Dialog box procedure for the clipboard editor. */
BOOL CALLBACK 
clip_edit_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg)  {
    case WM_INITDIALOG:
	SetWindowText (dlg, _("Clipboard Editor"));
	SetDlgItemText (dlg, IDC_CLIPEDIT_SEND, _("&Copy"));
	SetDlgItemText (dlg, IDC_CLIPEDIT_CLEAR, _("Clea&r"));
	SetDlgItemText (dlg, IDC_CLIPEDIT_LOAD, _("&Load"));	
	SetDlgItemText (dlg, IDC_CLIPEDIT_SAVE, _("&Save"));
	SetDlgItemText (dlg, IDC_CLIPEDIT_QUOTE, _("Add quotes")); 
	SetDlgItemText (dlg, IDOK, _("&Close"));

        load_clipboard (dlg);
	center_window (dlg, NULL);
	SetForegroundWindow (dlg);
	return TRUE;

    case WM_HELP:
	html_help_dispatch (lparam, "winpt.chm::winpt_texts.txt", help_arr);
	break;

    case WM_SYSCOMMAND:
	if (LOWORD (wparam) == SC_CLOSE)
	    EndDialog (dlg, TRUE);
	return TRUE;
        
    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDCANCEL:
	    EndDialog (dlg, FALSE);
	    return TRUE;

	case IDOK:		
	    EndDialog (dlg, TRUE);
	    return TRUE;
            
	case IDC_CLIPEDIT_SEND:
	    if (IsDlgButtonChecked (dlg, IDC_CLIPEDIT_QUOTE)) {
		gpgme_data_t txt;
		gpg_data_new_from_clipboard (&txt, 0);
		gpg_data_mail_quote (&txt);
		gpg_data_release_and_set_clipboard (txt, 0);
		SendMessage (dlg, WM_INITDIALOG, 0, 0);
		return TRUE;
	    }
	    SendDlgItemMessage (dlg, IDC_CLIPEDIT_CLIPTEXT2, EM_SETSEL, 0, -1);
	    SendDlgItemMessage (dlg, IDC_CLIPEDIT_CLIPTEXT2, WM_COPY, 0, 0);
	    return TRUE;
            
	case IDC_CLIPEDIT_CLEAR:
	    if (OpenClipboard (NULL) == FALSE) {
		msg_box (dlg,  winpt_strerror (WPTERR_CLIP_OPEN), 
			 _("Clipboard"), MB_ERR);
		return TRUE;
	    }			
	    if (EmptyClipboard () == FALSE)		
		msg_box (dlg, winpt_strerror (WPTERR_CLIP_EMPTY), 
			 _("Clipboard"), MB_ERR);
	    else
		SetDlgItemText (dlg, IDC_CLIPEDIT_CLIPTEXT2, "");
	    CloseClipboard ();
	    return TRUE;
            
	case IDC_CLIPEDIT_LOAD:
	    if (!load_clipboard_from_file (dlg)) {
		/* XXX: for some files the code produces a crash! */
		PostMessage (dlg, WM_COMMAND, MAKEWPARAM(IDC_CLIPEDIT_SEND, 0), 0);
	    }
	    return TRUE;
            
	case IDC_CLIPEDIT_SAVE:
	    save_clipboard_to_file (dlg);	    
	    return TRUE;
	}
	break;
    }
    
    return FALSE;
}
